'use strict';

var graphqlParseResolveInfo = require('graphql-parse-resolve-info');
var objection = require('objection');

objection.QueryBuilder.prototype.withGraphQL = function (info, options) {
    const resolve_tree = graphqlParseResolveInfo.parseResolveInfo(info);
    return process_resolve_tree_node({
        ...options,
        query: this,
        resolve_tree,
    });
};
function process_resolve_tree_node({ query, resolve_tree, filter, modifiers, relation, }) {
    var _a, _b, _c, _d, _e;
    const Model = query.modelClass();
    const type_name = Object.keys(resolve_tree.fieldsByTypeName)[0];
    const type_query_modifier = modifiers === null || modifiers === void 0 ? void 0 : modifiers[type_name];
    if (type_query_modifier) {
        type_query_modifier(query);
    }
    if (filter !== undefined && !isFilterObject(filter)) {
        throw Model.createValidationError({
            type: "InvalidFilter",
            message: `Invalid filter ${type_name}: must be an object.`,
        });
    }
    // Pull the list of getters from Model
    // see https://stackoverflow.com/a/39310917/189806
    const getters = new Set(Object.entries(Object.getOwnPropertyDescriptors(Model.prototype))
        .filter(([, descriptor]) => typeof descriptor.get === "function")
        .map(([key]) => key));
    // List of relation mappings
    // Static-cast them to RelationMappings, because thunk has already been resolved by now.
    const relations = Model.relationMappings;
    // Run select() and withGraphFetched() for fields requested in GraphQL query
    let has_selected_id = false;
    for (const resolve_subtree of Object.values(resolve_tree.fieldsByTypeName[type_name])) {
        const field = resolve_subtree.name;
        const modifier_name = `graphql.select.${field}`;
        const has_modifier = (_a = Model.modifiers) === null || _a === void 0 ? void 0 : _a[modifier_name];
        if (relations[field]) {
            // Nested relation
            query.withGraphFetched(`${field}(${field})`).modifiers({
                [field]: (subquery) => process_resolve_tree_node({
                    query: subquery,
                    resolve_tree: resolve_subtree,
                    filter: filter === null || filter === void 0 ? void 0 : filter[field],
                    modifiers,
                    relation: relations[field].relation,
                }),
            });
        }
        else if (getters.has(field) || has_modifier) ;
        else {
            // Normal field - select() it
            query.select(field);
            if (field === Model.idColumn) {
                has_selected_id = true;
            }
        }
        // Call field modifier
        if (has_modifier) {
            query.modify(modifier_name);
        }
    }
    if (!has_selected_id) {
        // Always select ID:
        // 1. This is useful for potential $query()
        // 2. This avoid automatic "select *" when not a single normal field is selected
        query.select(Model.idColumn);
    }
    // Add where() clauses based on the provided filter object
    if (filter) {
        for (const [field, value] of Object.entries(filter)) {
            if (value === undefined) {
                throw Model.createValidationError({
                    type: "InvalidFilter",
                    message: `Invalid filter ${type_name}.${field}: undefined.`,
                });
            }
            if (relations[field]) ;
            else if ((_b = Model.modifiers) === null || _b === void 0 ? void 0 : _b[field]) {
                // Call modifier
                query.modify(field, value);
            }
            else {
                // Normal filter
                if (isFilterObject(value)) {
                    throw Model.createValidationError({
                        type: "InvalidFilter",
                        message: `Invalid filter ${type_name}.${field}: must not be an object.`,
                    });
                }
                const [dbfield, op] = field.split("__");
                if (!op || op === "exact") {
                    if (Array.isArray(value)) {
                        throw Model.createValidationError({
                            type: "InvalidFilter",
                            message: `Invalid filter ${type_name}.${field}: must not be an array.`,
                        });
                    }
                    if (value !== null) {
                        query.where(dbfield, value);
                    }
                    else {
                        query.whereNull(dbfield);
                    }
                }
                else if (op === "in") {
                    if (!Array.isArray(value)) {
                        throw Model.createValidationError({
                            type: "InvalidFilter",
                            message: `Invalid filter ${type_name}.${field}: must be an array.`,
                        });
                    }
                    query.whereIn(dbfield, value);
                }
                else {
                    throw Model.createValidationError({
                        type: "InvalidFilter",
                        message: `Invalid filter ${type_name}.${field}: unsupported operator ${op}.`,
                    });
                }
            }
        }
    }
    if ((_c = Model.modifiers) === null || _c === void 0 ? void 0 : _c.graphql) {
        query.modify("graphql");
    }
    if (((_d = Model.modifiers) === null || _d === void 0 ? void 0 : _d["graphql.many"]) && !query.has("first")) {
        // TODO: exclude XxxOne relations?
        query.modify("graphql.many");
    }
    if (!relation && ((_e = Model.modifiers) === null || _e === void 0 ? void 0 : _e["graphql.top"])) {
        query.modify("graphql.top");
    }
    return query;
}
function isFilterObject(obj) {
    return typeof obj === "object" && !Array.isArray(obj);
}
//# sourceMappingURL=index.js.map
